const uint32_t ay_channel_remap_table[6 * 16] = {
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, //LAY_ABC
  0, 1, 4, 5, 2, 3, 6, 7, 8, 10, 9, 11, 12, 13, 14, 15, //LAY_ACB
  2, 3, 0, 1, 4, 5, 6, 7, 9, 8, 10, 11, 12, 13, 14, 15, //LAY_BAC
  2, 3, 4, 5, 0, 1, 6, 7, 9, 10, 9, 11, 12, 13, 14, 15, //LAY_BCA
  4, 5, 0, 1, 2, 3, 6, 7, 10, 8, 9, 11, 12, 13, 14, 15, //LAY_CAB
  4, 5, 2, 3, 0, 1, 6, 7, 10, 9, 8, 11, 12, 13, 14, 15, //LAY_CBA
};

const uint32_t ay_mixer_remap_table[6 * 3] = {
  0, 1, 2, //LAY_ABC
  0, 2, 1, //LAY_ACB,
  1, 0, 2, //LAY_BAC,
  1, 2, 0, //LAY_BCA,
  2, 0, 1, //LAY_CAB,
  2, 1, 0, //LAY_CBA
};

uint8_t hway_reg_1[16];
uint8_t hway_reg_2[16];



void hway_set_clock(uint32_t f)
{
  i2s_set_rate(f / 32);
}



void hway_write(uint8_t chip, uint8_t reg, uint8_t val)
{
  uint8_t bdir = !chip ? AY1_MCP_PIN_BDIR : AY2_MCP_PIN_BDIR;

  if (!chip) hway_reg_1[reg] = val; else hway_reg_2[reg] = val;

  mcp.digitalWrite(bdir, LOW); //inactive
  mcp.digitalWrite(AY_MCP_PIN_BC1, LOW);
  delayMicroseconds(5);
  mcp.writePort(1, reg); //reg number to GPB
  //delayMicroseconds(5);
  mcp.digitalWrite(AY_MCP_PIN_BC1, HIGH);
  mcp.digitalWrite(bdir, HIGH);
  delayMicroseconds(5);
  mcp.digitalWrite(bdir, LOW);
  mcp.digitalWrite(AY_MCP_PIN_BC1, LOW);
  delayMicroseconds(5);
  mcp.writePort(1, val);  //reg data to GPB
  //delayMicroseconds(5);
  mcp.digitalWrite(bdir, HIGH);
  delayMicroseconds(5);
  mcp.digitalWrite(bdir, LOW);
}


void hway_write_remap(uint8_t chip, uint8_t reg, uint8_t val)
{
  uint8_t old;
  uint32_t off;

  if (reg >= 16) return;

  if (reg == 0x07) //mixer bits remapping
  {
    off = Config.ay_layout * 3;

    old = val;

    val = old & 0xc0;
    val |= ((old >> 0) & 0x09) << ay_mixer_remap_table[off + 0];
    val |= ((old >> 1) & 0x09) << ay_mixer_remap_table[off + 1];
    val |= ((old >> 2) & 0x09) << ay_mixer_remap_table[off + 2];
  }
  else
  {
    reg = ay_channel_remap_table[Config.ay_layout * 16 + reg];
  }

  hway_write(chip, reg, val);
}



void hway_reset(void)
{
  mcp.digitalWrite(AY_MCP_PIN_RESET, LOW);
  delayMicroseconds(5);
  mcp.digitalWrite(AY_MCP_PIN_RESET, HIGH);
  delayMicroseconds(5);

  for (int i = 0; i < 16; ++i)
  {
    hway_write(0, i, 0);
    hway_write(1, i, 0);
  }
}
