#define PAD_FWD           0x0010
#define PAD_PLAY          0x0020
#define PAD_REW           0x0040
#define PAD_STOP          0x0080
#define PAD_ANY           (PAD_REW|PAD_PLAY|PAD_FWD|PAD_STOP)

#define PAD_FIRST_REPEAT  16
#define PAD_SECOND_REPEAT 1


uint16_t pad_state;
uint16_t pad_state_prev;
uint16_t pad_state_down;
uint16_t pad_state_up;
uint16_t pad_state_rep;
uint8_t pad_time;
uint8_t pad_locked;
uint8_t pad_unlock_cnt;
uint16_t pad_lock_prev;
uint8_t pad_wait_release;
uint8_t pad_redraw_needed;


void pad_init()
{
  pad_state = 0;
  pad_state_prev = 0;
  pad_state_down = 0;
  pad_state_up = 0;
  pad_state_rep = 0;
  pad_time = 0;
  pad_locked = 0;
  pad_unlock_cnt = 0;
  pad_lock_prev = 0;
  pad_wait_release = 0;
  pad_redraw_needed = 0;
}



void pad_lock(bool lock)
{
  pad_locked = lock ? 1 : 0;
  pad_unlock_cnt = 0;
  pad_lock_prev = 0;

  if (!lock)
  {
    pad_wait_release = 1;
    pad_redraw_needed = 1;

    frame_set_display_timeout(0);
  }
}



bool pad_is_locked()
{
  return pad_locked ? true : false;
}



bool pad_force_redraw()
{
  return pad_redraw_needed ? true : false;
}



int pad_check()
{
  uint16_t state = (mcp.readPort(0) & 0xf0) ^ 0xf0;

  pad_redraw_needed = 0;

  if (state & PAD_ANY)
  {
    if (!display_is_active())
    {
      if (pad_is_locked()) frame_set_display_timeout(1);

      display_enable(true);
    }
    else
    {
      if (!pad_is_locked()) frame_set_display_timeout(0);
    }
  }

  if (pad_is_locked())
  {
    int down = state ^ pad_lock_prev & state;

    pad_lock_prev = state;

    if (down)
    {
      frame_set_display_timeout(1);

      ++pad_unlock_cnt;

      if (pad_unlock_cnt >= 5)
      {
        pad_lock(false);
      }
    }

    return 0;
  }

  if (pad_wait_release)
  {
    if (state) return 0;

    pad_wait_release = 0;
  }

  pad_state_prev = pad_state;
  pad_state = state;
  pad_state_down = pad_state ^ pad_state_prev & pad_state;
  pad_state_up = ~pad_state ^ ~pad_state_prev & ~pad_state;
  pad_state_rep = pad_state_down;

  if (pad_state_down)
  {
    pad_time = PAD_FIRST_REPEAT;
  }

  if (pad_state)
  {
    if (pad_time)
    {
      --pad_time;

      if (!pad_time)
      {
        pad_time = PAD_SECOND_REPEAT;

        pad_state_rep |= pad_state;
      }
    }
  }
  else
  {
    pad_time = 0;
  }

  return pad_state;
}



//0 no timeout, otherwise timeout in ms

void pad_wait_any(int timeout)
{
  timeout /= 100;

  while (1)
  {
    pad_check();

    if (pad_state_down & PAD_ANY) break;

    if (timeout)
    {
      --timeout;

      if (timeout <= 0) break;
    }

    delay(1000 / 50);
  }
}
