struct AYSongInfo AYInfo;

uint8_t music_type;

#define INFO_DELAY      32
#define REW_HOLD_TIME   16
#define FWD_HOLD_TIME   16


//subsong is only used for AY format

int music_open(const char* filename, int ay_sub_song)
{
  int err = FILE_ERR_OTHER;

  if (!sd_fat.begin(SD_CS, SD_SPI_SPEED)) return FILE_ERR_NO_CARD;

  if (sd_file.open(filename, O_RDONLY))
  {
    sd_file.getName(lfn, sizeof(lfn));

    music_type = browser_check_ext(lfn);

    if (music_type != TYPE_AY)
    {
      if (music_type != TYPE_UNK && music_type != TYPE_AYL)
      {
        music_data_size = sd_file.fileSize();

        if (music_data_size <= sizeof(music_data))
        {
          sd_file.read((char*)music_data, music_data_size);
          err = FILE_ERR_NONE;
        }
        else
        {
          err = FILE_ERR_TOO_LARGE;
        }
      }
    }
    else
    {
      err = AY_Load(AYInfo, ay_sub_song); //uses currently opened file stream
    }

    sd_file.close();
  }

  return err;
}



void music_play()
{
  memset(&AYInfo, 0, sizeof(AYInfo));

  AYInfo.module = music_data;
  AYInfo.module_len = music_data_size;
  AYInfo.file_data = music_data;
  AYInfo.file_len = music_data_size;
  /*
    AYInfo.data = NULL; //needs to be reset to prevent player deleting arrays
    AYInfo.data1 = NULL;
    AYInfo.module1 = NULL;
    AYInfo.Length = 0;
    AYInfo.Loop = 0;
    AYInfo.is_ts = 0;

    AYInfo.Author.reserve(64);
    AYInfo.Author = ""; //in case info is missing, but there is some from previous track
    AYInfo.Name.reserve(64);
    AYInfo.Name = "";
  */
  switch (music_type)
  {
    case TYPE_PT1: PT1_Init(AYInfo); PT1_GetInfo(AYInfo); break;
    case TYPE_PT2: PT2_Init(AYInfo); PT2_GetInfo(AYInfo); break;
    case TYPE_PT3: PT3_Init(AYInfo); PT3_GetInfo(AYInfo); break;
    case TYPE_STC: STC_Init(AYInfo); STC_GetInfo(AYInfo); break;
    case TYPE_STP: STP_Init(AYInfo); STP_GetInfo(AYInfo); break;
    case TYPE_ASC: ASC_Init(AYInfo); ASC_GetInfo(AYInfo); break;
    case TYPE_PSC: PSC_Init(AYInfo); PSC_GetInfo(AYInfo); break;
    case TYPE_SQT: SQT_Init(AYInfo); SQT_GetInfo(AYInfo); break;
    case TYPE_AY: AY_Init(AYInfo); AY_GetInfo(AYInfo); break;
  }
}



void music_update(void)
{
  switch (music_type)
  {
    case TYPE_PT1: PT1_Play(AYInfo); break;
    case TYPE_PT2: PT2_Play(AYInfo); break;
    case TYPE_PT3: PT3_Play(AYInfo); break;
    case TYPE_STC: STC_Play(AYInfo); break;
    case TYPE_STP: STP_Play(AYInfo); break;
    case TYPE_ASC: ASC_Play(AYInfo); break;
    case TYPE_PSC: PSC_Play(AYInfo); break;
    case TYPE_SQT: SQT_Play(AYInfo); break;
    case TYPE_AY: break; //AY files play differently, they're running Z80 emulation tied to the sound ISR
  }
}


void music_stop()
{
  hway_reset();

  switch (music_type)
  {
    case TYPE_PT1: PT1_Cleanup(AYInfo); break;
    case TYPE_PT2: PT2_Cleanup(AYInfo); break;
    case TYPE_PT3: PT3_Cleanup(AYInfo); break;
    case TYPE_STC: STC_Cleanup(AYInfo); break;
    case TYPE_STP: STP_Cleanup(AYInfo); break;
    case TYPE_ASC: ASC_Cleanup(AYInfo); break;
    case TYPE_PSC: PSC_Cleanup(AYInfo); break;
    case TYPE_SQT: SQT_Cleanup(AYInfo); break;
    case TYPE_AY: AY_Cleanup(AYInfo); break;
  }
}



void player_draw_header(PGM_P str)
{
  display_mode(DM_AND);
  display_rect(0, 0, 110, 8);
  display_mode(DM_OR);
  display_str_P(4, 0, str);
}



//return true if another track has to be played, false when playing is done

bool player_screen(const char* filename_)
{
  char info_str[128], temp[64];
  char filename[MAX_PATH];
  int ay_cur_song = 0;
  bool ay_reload = true;
  bool play_more = true;
  int rew_down = 0;
  int fwd_down = 0;

  strncpy(filename, filename_, sizeof(filename) - 1); //contents of filename gets truncated on second loop by some reason, while this copy remains intact

  while (ay_reload)
  {
    ay_reload = false;

    switch (music_open(filename, ay_cur_song))
    {
      case FILE_ERR_NO_CARD:            message_box_P(PSTR(MSG_ERR_NO_CARD), MB_WARN); browser_reset_directory(); return false;
      case FILE_ERR_TOO_LARGE:          message_box_P(PSTR(MSG_ERR_TOO_LARGE), 50 * 2); return false; //these three messages displayed for 2 seconds without need to pressing button
      case FILE_ERR_UNK_FORMAT:         message_box_P(PSTR(MSG_ERR_UNK_FORMAT), 50 * 2); return false;
      case FILE_ERR_NOT_ENOUGH_MEMORY:  message_box_P(PSTR(MSG_ERR_NOT_ENOUGH_MEMORY), 50 * 2); return false;
      case FILE_ERR_OTHER:              message_box_P(PSTR(MSG_ERR_CANT_LOAD_FILE), MB_WARN); return false;
    }

    music_play();

    spec_init();

    display_clear();

    memset(info_str, 0, sizeof(info_str));

    if (strlen(AYInfo.Author) > 0) strncat(info_str, AYInfo.Author, sizeof(info_str) - 1);

    if ((strlen(AYInfo.Name) > 0) && (strlen(AYInfo.Author) > 0))
    {
      strncat(info_str, " - ", sizeof(info_str) - 1);
    }

    if (strlen(AYInfo.Name) > 0) strncat(info_str, AYInfo.Name, sizeof(info_str) - 1);

    int frame_play = 0;
    int frame_step = 1;
    int frame = 0;

    bool pause = false;
    bool update_static = true;

    bool slow = false;

    hway_set_clock(Config.ay_clock);
    hway_reset();
    frame_rate_set(Config.ay_fps);

    int info_sx = 0;
    int info_wdt = display_str_len(info_str);
    int info_dir = -1;
    int info_delay = INFO_DELAY;

    bool loop = false;

    if (info_wdt <= VIEW_WDT) info_dir = 0;

    while (true)
    {
      if (pad_force_redraw()) update_static = true;

      if (display_is_active())
      {
        if (update_static)
        {
          update_static = false;

          player_draw_header(!pause ? PSTR("\4 Now playing...") : PSTR("\7 Paused"));

          display_rect(0, 8, VIEW_WDT, 1);

          int sx = SPEC_SX;
          int sy = SPEC_SY;

          for (int i = 0; i < SPEC_BANDS; ++i)
          {
            display_rect(sx, SPEC_SY + SPEC_HEIGHT + 1, SPEC_BAND_WIDTH - 1, 1);

            sx += SPEC_BAND_WIDTH;
          }

          switch (music_type)
          {
            case TYPE_PT1: display_str_P(0, VIEW_HGT - 7, PSTR("pt1")); break;
            case TYPE_PT2: display_str_P(0, VIEW_HGT - 7, PSTR("pt2")); break;
            case TYPE_PT3: display_str_P(0, VIEW_HGT - 7, PSTR("pt3")); break;
            case TYPE_STC: display_str_P(0, VIEW_HGT - 7, PSTR("stc")); break;
            case TYPE_STP: display_str_P(0, VIEW_HGT - 7, PSTR("stp")); break;
            case TYPE_ASC: display_str_P(0, VIEW_HGT - 7, PSTR("asc")); break;
            case TYPE_PSC: display_str_P(0, VIEW_HGT - 7, PSTR("psc")); break;
            case TYPE_SQT: display_str_P(0, VIEW_HGT - 7, PSTR("sqt")); break;
            case TYPE_AY:
              {
                if (AY_GetNumSongs() == 0)
                {
                  strcpy(temp, "ay");
                }
                else
                {
                  sprintf(temp, "ay %u/%u", ay_cur_song + 1, AY_GetNumSongs() + 1);
                }

                display_str(0, VIEW_HGT - 7, temp);
              }
              break;
          }
        }

        display_charge(111, 0);

        display_mode(DM_AND);
        display_rect(0, 12, VIEW_WDT, 8);
        display_mode(DM_OR);

        display_str(info_sx, 12, info_str);

        int sx = SPEC_SX;
        int sy = SPEC_SY;

        display_mode(DM_AND);
        display_rect(0, sy, VIEW_WDT, SPEC_HEIGHT);
        display_mode(DM_OR);

        for (int i = 0; i < SPEC_BANDS; ++i)
        {
          int h = spec_levels[i];

          if (h > SPEC_HEIGHT) h = SPEC_HEIGHT;

          display_mask(spec_colors[i]);
          display_rect(sx, sy + SPEC_HEIGHT - h, SPEC_BAND_WIDTH - 1, h);

          sx += SPEC_BAND_WIDTH;
        }

        display_mask(0);

        if (AYInfo.Length >= 0)
        {
          sprintf(temp, "%2.2u:%2.2u/%2.2u:%2.2u", frame_play / 50 / 60, frame_play / 50 % 60, AYInfo.Length / 50 / 60, AYInfo.Length / 50 % 60);
        }
        else
        {
          sprintf(temp, "%2.2u:%2.2u/%c", frame_play / 50 / 60, frame_play / 50 % 60, 8);
        }

        sx = VIEW_WDT - display_str_len(temp);
        sy = VIEW_HGT - 7;

        display_mode(DM_AND);
        display_rect(sx - 8, sy, VIEW_WDT - sx + 8, VIEW_HGT - sy);
        display_mode(DM_OR);
        display_str(sx, sy, temp);
      }

      display_update();

      if (!pause)
      {
        spec_update(AYInfo.is_ts);

        int f = frame_step;

        if (f <= 1 && slow) f = frame & 1;

        for (int i = 0; i < f; ++i)
        {
          music_update();

          ++frame_play;

          if (AYInfo.Length >= 0) //when the length is known (not infinite)
          {
            if (frame_play >= AYInfo.Length)
            {
              loop = true;
              frame_play -= AYInfo.Length;
            }
          }

          if (frame_step > 1)
          {
            yield();
          }
        }
      }

      if (music_type == TYPE_AY)  //AY format syncs up by the sound buffer
      {
        while (!Sound.buf_do_update)
        {
          yield();
        }

        Sound.buf_do_update = false;

        if (!pause)
        {
          AY_PlayBuf();

          spec_add(500, (Sound.buf_rd[0] + Sound.buf_rd[SOUND_BUF_MAX_SIZE / 4] + Sound.buf_rd[SOUND_BUF_MAX_SIZE / 2] + Sound.buf_rd[SOUND_BUF_MAX_SIZE / 2 + SOUND_BUF_MAX_SIZE / 4]) / 4, SPEC_ENV_COL);
        }
      }
      else
      {
        wait_frame();
      }
      /*
            Serial.println(Sound.int_counter);
            Sound.int_counter=0;
      */
      delay(0);

      pad_check();

      if (loop)
      {
        if (Config.file_mode == FILE_MODE_ONCE)
        {
          pad_state_up |= PAD_FWD; //force track change
          fwd_down = 0;
        }
      }

      if (pad_state_down & PAD_PLAY)
      {
        pause ^= true;
        update_static = true;

        if (pause)
        {
          //mute sound

          sound_clear_buf();

          for (uint8_t chip = 0; chip < 2; ++chip)
          {
            hway_write(chip, 0x07, 0xff);
            hway_write(chip, 0x08, 0x00);
            hway_write(chip, 0x09, 0x00);
            hway_write(chip, 0x0a, 0x00);
          }
        }
      }

      if (pad_state & PAD_REW)
      {
        ++rew_down;

        if (rew_down > (REW_HOLD_TIME + 1)) rew_down = (REW_HOLD_TIME + 1); //to avoid skips in multisong AY files

        if (rew_down == REW_HOLD_TIME)
        {
          if (music_type != TYPE_AY) //slowdown for regular music formats, switch to a prev subsong for the AY format
          {
            slow = true;
          }
          else
          {
            if (ay_cur_song > 0)
            {
              --ay_cur_song;
              ay_reload = true;

              player_draw_header(PSTR("\3 PrevSub"));

              break;
            }
          }
        }
      }
      else
      {
        if (rew_down < REW_HOLD_TIME)
        {
          if (pad_state_up & PAD_REW)
          {
            player_draw_header(PSTR("\3 Prev"));

            if (Config.dir_mode != DIR_MODE_SHUFFLE)
            {
              if (browser_move_cur(-1, Config.dir_mode == DIR_MODE_LOOP ? true : false))
              {
                play_more = false;
              }
            }
            else
            {
              browser_shuffle_cur();
            }

            break;
          }
        }

        rew_down = 0;
        slow = false;
      }

      if (pad_state & PAD_FWD)
      {
        ++fwd_down;

        if (fwd_down > (FWD_HOLD_TIME + 1)) fwd_down = (FWD_HOLD_TIME + 1); //to avoid skips in multisong AY files

        if (fwd_down == FWD_HOLD_TIME)
        {
          if (music_type != TYPE_AY) //fast forward for regular music formats, switch to a next subsong for the AY format
          {
            frame_step = 60;
          }
          else
          {
            if (ay_cur_song < AY_GetNumSongs())
            {
              ++ay_cur_song;
              ay_reload = true;

              player_draw_header(PSTR("\5 NextSub"));
              break;
            }
          }
        }
      }
      else
      {
        if (fwd_down < FWD_HOLD_TIME)
        {
          if (pad_state_up & PAD_FWD)
          {
            player_draw_header(PSTR("\5 Next"));

            if (Config.dir_mode != DIR_MODE_SHUFFLE)
            {
              if (browser_move_cur(1, Config.dir_mode == DIR_MODE_LOOP ? true : false))
              {
                play_more = false;
              }
            }
            else
            {
              browser_shuffle_cur();
            }

            break;
          }
        }

        fwd_down = 0;
        frame_step = 1;
      }

      if (pad_state_up & PAD_STOP)
      {
        play_more = false;
        break;
      }

      if (info_delay)
      {
        --info_delay;
      }
      else
      {
        info_sx += info_dir;

        if (info_dir < 0)
        {
          if (info_sx + info_wdt <= VIEW_WDT)
          {
            info_dir = 1;
            info_delay = INFO_DELAY;
          }
        }
        else if (info_dir > 0)
        {
          if (info_sx >= 0)
          {
            info_dir = -1;
            info_delay = INFO_DELAY;
          }
        }
      }

      ++frame;

      if ((music_type == TYPE_AY) && AY_IsMemFail()) break;
    }

    music_stop();
    sound_clear_buf();

    if ((music_type == TYPE_AY) && AY_IsMemFail())
    {
      message_box_P(PSTR(MSG_ERR_NOT_ENOUGH_MEMORY), MB_WARN);
      return false;
    }

    if (play_more)
    {
      display_update();
      delay(1000 / 5);
    }
  }

  return play_more;
}
