enum {
  MB_WARN = 0,
  MB_OK,
  MB_YESNO,
  MB_TIMEOUT
};



bool message_box_u(const char* msg, int type, bool pgm)
{
  char c;
  bool ret = true;
  int sel_delay = 0;
  int h = (type >= MB_TIMEOUT) ? 64 : 57;

  bool pad_was_down = (pad_state & PAD_ANY) ? true : false;

  while (1)
  {
    if (!(frame_cnt & 15) || sel_delay)
    {
      display_clear();

      display_mask(2);
      display_rect(0, 0, 128, h);
      display_mask(0);
      display_mode(DM_AND);
      display_rect(1, 1, 128 - 2, h - 2);
      display_mode(DM_OR);

      if ((frame_cnt & 16) || type == MB_OK || type >= MB_TIMEOUT)
      {
        int ptr = 0;
        int lines = 1;

        while (1)
        {
          if (!pgm)
          {
            c = msg[ptr++];
          }
          else
          {
            c = pgm_read_byte(&msg[ptr++]);
          }

          if (!c) break;
          if (c == '\n') ++lines;
        }

        int sy = h / 2 - lines * 8 / 2;

        ptr = 0;

        while (1)
        {
          if (!pgm)
          {
            if (!msg[ptr]) break;

            display_str(128 / 2 - display_str_len(&msg[ptr]) / 2, sy, &msg[ptr]);
          }
          else
          {
            if (!pgm_read_byte(&msg[ptr])) break;

            display_str_P(128 / 2 - display_str_len_P(&msg[ptr]) / 2, sy, &msg[ptr]);
          }

          sy += 8;

          while (1)
          {
            if (!pgm)
            {
              c = msg[ptr];
            }
            else
            {
              c = pgm_read_byte(&msg[ptr]);
            }

            if (!c) break;

            ++ptr;

            if (c == '\n') break;
          }
        }
      }

      if (type < MB_TIMEOUT)
      {
        if (type == MB_OK || type == MB_WARN || type == MB_YESNO)
        {
          if (!(sel_delay & 4) || !ret) display_str_P(108, 58, PSTR("\4OK"));
        }
        if (type == MB_YESNO)
        {
          if (!(sel_delay & 4) || ret) display_str_P(0, 58, PSTR("\6CANCEL"));
        }
      }

      display_update();
    }

    if (sel_delay)
    {
      --sel_delay;

      if (!sel_delay) break;
    }
    else
    {
      if (type < MB_TIMEOUT)
      {
        pad_check();

        if (pad_was_down)
        {
          if (!(pad_state & PAD_ANY)) pad_was_down = false;
        }
        else
        {
          if (type == MB_YESNO)
          {
            if (pad_state_up & PAD_STOP)
            {
              sound_play(SFX_CANCEL);
              ret = false;
              sel_delay = 8;
            }
          }

          if (pad_state_up & PAD_PLAY)
          {
            sound_play(SFX_SELECT);
            ret = true;
            sel_delay = 8;
          }
        }
      }
      else
      {
        --type;

        if (type <= MB_TIMEOUT) break;
      }
    }

    wait_frame();
  }

  return ret;
}



bool message_box(const char* msg, int type)
{
  return message_box_u(msg, type, false);
}



bool message_box_P(PGM_P msg, int type)
{
  return message_box_u((const char*)msg, type, true);
}



void wait_screen()
{
  char buf[16];
  static uint8_t flash = 0;

  strcpy(buf, MSG_WAIT_SCREEN);

  buf[0] = 16 + flash;

  flash = (flash + 1) & 3;

  display_clear();
  display_str(128 / 2 - display_str_len(buf) / 2, 32 - 4, buf);
  display_update();
}



void display_num_small(int sx, int sy, int n)
{
  if (n < 10) sx -= 5 / 2; else if (n < 100) sx -= 10 / 2; else sx -= 15 / 2;

  if (n >= 100)
  {
    display_char(sx, sy, 0x16 + n / 100 % 10);
    sx += 5;
  }

  if (n >= 10)
  {
    display_char(sx, sy, 0x16 + n / 10 % 10);
    sx += 5;
  }

  display_char(sx, sy, 0x16 + n % 10);
}



void display_charge(int sx, int sy)
{
  int charge = battery_level();

#ifndef DEBUG_BATTERY
  display_rect(sx, sy + 2, 2, 3);
  display_rect(sx + 2, sy, 15, 7);
  display_mode(DM_AND);

  if (charge <= 100)
  {
    if (charge >= 10 && charge < 100) ++sx;

    display_num_small(sx + 2 + 15 / 2 , sy + 1, charge);
  }
  else
  {
    if (frame_cnt & 32) display_char(sx + 2 + 4, sy + 1, 0x14);
  }
#else
  display_num_small(sx + 2 + 15 / 2 , sy + 1, charge);
#endif

  display_mode(DM_OR);
}
